// --------------------------------------------------------------------------------------------------------------------
// <copyright file="WavInfo.cs" company="QutEcoacoustics">
// All code in this file and all associated files are the copyright and property of the QUT Ecoacoustics Research Group (formerly MQUTeR, and formerly QUT Bioacoustics Research Group).
// </copyright>
// <summary>
//   Wav Info.
// </summary>
// --------------------------------------------------------------------------------------------------------------------

namespace AudioAnalysisTools.DSP
{
    using System;
    using System.Linq;

    /// <summary>
    /// Wav Info.
    /// </summary>
    public class WavInfo
    {
        /// <summary>
        /// Gets or sets Number of channels.
        /// </summary>
        public short Channels { get; set; }

        /// <summary>
        /// Gets or sets Sample rate of audio.
        /// </summary>
        public int SampleRate { get; set; }

        /// <summary>
        /// Gets or sets Bits per sample.
        /// </summary>
        public short BitsPerSample { get; set; }

        /// <summary>
        /// Gets or sets bytes Per Sample / Block Align.
        /// </summary>
        public short BytesPerSample { get; set; }

        /// <summary>
        /// Gets or sets.
        /// </summary>
        public short CompressionCode { get; set; }

        /// <summary>
        /// Gets or sets.
        /// </summary>
        public int BytesPerSecond { get; set; }

        /// <summary>
        /// Gets or sets.
        /// </summary>
        public double[] Samples { get; set; }

        /// <summary>
        /// Gets or sets.
        /// </summary>
        public short[][] SamplesSplit { get; set; }

        /// <summary>
        /// Gets or sets.
        /// </summary>
        public TimeSpan Duration { get; set; }

        /// <summary>
        /// Gets.
        /// </summary>
        public double Epsilon => Math.Pow(0.5, this.BitsPerSample - 1);

        public int Frames { get; set; }

        /// <summary>
        /// Initializes a new instance of the <see cref="WavInfo"/> class.
        /// Create a sine wave.
        /// </summary>
        public WavInfo(double freq, double amp, double phase, TimeSpan duration, int sampleRate)
        {
            int length = (int)Math.Floor(duration.TotalSeconds * sampleRate);
            double[] data = new double[length];

            for (int i = 0; i < length; i++)
            {
                data[i] = amp * Math.Sin(phase + (2.0 * Math.PI * freq * i / sampleRate));
            }

            this.Channels = 1;
            this.BitsPerSample = 16;
            this.SampleRate = sampleRate;
            this.Samples = data;
            this.Duration = duration;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WavInfo"/> class.
        /// Create a Sine wav with multiple frequencies.
        /// </summary>
        public WavInfo(double[] freq, double amp, double phase, TimeSpan duration, int sampleRate)
        {
            int length = (int)Math.Floor(duration.TotalSeconds * sampleRate);
            double[] data = new double[length];

            int count = freq.Length;

            for (int i = 0; i < length; i++)
            {
                for (int f = 0; f < count; f++)
                {
                    data[i] += amp * Math.Sin(phase + (2.0 * Math.PI * freq[f] * i / sampleRate));
                }
            }

            this.Channels = 1;
            this.BitsPerSample = 16;
            this.SampleRate = sampleRate;
            this.Samples = data;
            this.Duration = duration;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="WavInfo"/> class.
        /// Create a wav info using data, channels, bits per sample and sample rate.
        /// </summary>
        /// <param name="rawData">Raw samples.</param>
        /// <param name="channels">Number of channels.</param>
        /// <param name="bitsPerSample">Bits per sample.</param>
        /// <param name="sampleRate">Sample rate.</param>
        public WavInfo(double[] rawData, short channels, short bitsPerSample, int sampleRate)
        {
            this.Channels = channels;
            this.BitsPerSample = bitsPerSample;
            this.SampleRate = sampleRate;
            this.Samples = rawData;
        }

        public WavInfo()
        {
        }

        /// <summary>
        /// Subsamples audio.
        /// </summary>
        /// <param name="interval">
        /// Keeps every <paramref name="interval"/> sample.
        /// </param>
        /// <param name="wavInfo">Wav info.</param>
        public static WavInfo SubSample(WavInfo wavInfo, int interval)
        {
            if (interval <= 1)
            {
                // no changes required.
                return wavInfo;
            }

            int length = wavInfo.Samples.Length;
            int newLength = length / interval;
            double[] newSamples = new double[newLength];

            //want length to be exact mulitple of interval
            length = newLength * interval;

            // copy only required samples
            for (int i = 0; i < newLength; i++)
            {
                newSamples[i] = wavInfo.Samples[i * interval];
            }

            return new WavInfo(newSamples, wavInfo.Channels, wavInfo.BitsPerSample, wavInfo.SampleRate / interval);
        }

        /// <summary>
        ///
        /// </summary>
        public static double CalculateMaximumAmplitude(WavInfo wavInfo)
        {
            // Max expects at least 1 item in array.
            if (wavInfo.Samples.Length > 0)
            {
                return wavInfo.Samples.Max();
            }

            return 0;
        }
    }
}